"""Utility functions for pyadtpulse."""

import logging
import string
import sys
from base64 import urlsafe_b64encode
from datetime import datetime, timedelta
from pathlib import Path
from random import randint
from threading import RLock, current_thread

from bs4 import BeautifulSoup
from yarl import URL

LOG = logging.getLogger(__name__)


def remove_prefix(text: str, prefix: str) -> str:
    """Remove prefix from a string.

    Args:
        text (str): original text
        prefix (str): prefix to remove

    Returns:
        str: modified string
    """
    return text[text.startswith(prefix) and len(prefix) :]


def handle_response(code: int, url: URL | None, level: int, error_message: str) -> bool:
    """Handle the response from query().

    Args:
        code (int): the return code
        level (int): Level to log on error (i.e. INFO, DEBUG)
        error_message (str): the error message

    Returns:
        bool: True if no error occurred.
    """
    if code >= 400:
        LOG.log(level, "%s: error code = %s from %s", error_message, code, url)
        return False
    return True


def make_soup(
    code: int,
    response_text: str | None,
    url: URL | None,
    level: int,
    error_message: str,
) -> BeautifulSoup | None:
    """Make a BS object from a Response.

    Args:
        code (int): the return code
        response_text Optional(str): the response text
        level (int): the logging level on error
        error_message (str): the error message

    Returns:
        Optional[BeautifulSoup]: a BS object, or None on failure
    """
    if not handle_response(code, url, level, error_message):
        return None
    if response_text is None:
        LOG.log(level, "%s: no response received from %s", error_message, url)
        return None
    return BeautifulSoup(response_text, "html.parser")


FINGERPRINT_LENGTH = 2292
ALLOWABLE_CHARACTERS = list(string.ascii_letters + string.digits)
FINGERPRINT_RANGE_LEN = len(ALLOWABLE_CHARACTERS)


def generate_random_fingerprint() -> str:
    """Generate a random browser fingerprint string.

    Returns:
        str: a fingerprint string
    """
    fingerprint = [
        ALLOWABLE_CHARACTERS[(randint(0, FINGERPRINT_RANGE_LEN - 1))]
        for i in range(FINGERPRINT_LENGTH)
    ]
    return "".join(fingerprint)


def generate_fingerprint_from_browser_json(filename: str) -> str:
    """Generate a browser fingerprint from a JSON file.

    Args:
        filename (str): JSON file containing fingerprint information

    Returns:
        str: the fingerprint
    """
    data = Path(filename).read_text(encoding="utf-8")
    # Pulse just calls JSON.Stringify() and btoa() in javascript, so we need to
    # do this to emulate that
    data2 = "".join(data.split())
    return str(urlsafe_b64encode(data2.encode("utf-8")), "utf-8")


class DebugRLock:
    """Provides a debug lock logging caller who acquired/released."""

    def __init__(self, name: str):
        """Create the lock."""
        self._Rlock = RLock()
        self._lock_name = name

    def acquire(self, blocking: bool = True, timeout: float = -1) -> bool:
        """Acquire the lock.

        Args:
            blocking (bool, optional): blocks if can't obtain the lock if True.
                                       Defaults to True.
            timeout (float, optional): timeout to wait to acquire the lock.
                                       Defaults to -1.

        Returns:
            bool: True if lock obtained, False if blocking is False and lock couldn't be
                  obtained
        """
        caller = sys._getframe().f_back
        thread_name = current_thread().name
        if caller is not None:
            caller2 = caller.f_code.co_name
        else:
            caller2 = "*Unknown*"
        LOG.debug(
            "acquiring lock %s blocking: %s from %s from thread %s",
            self._lock_name,
            blocking,
            caller2,
            thread_name,
        )
        retval = self._Rlock.acquire(blocking, timeout)
        LOG.debug(
            "acquisition of %s from %s from thread %s  returned %d info: %s",
            self._lock_name,
            caller2,
            thread_name,
            retval,
            repr(self._Rlock),
        )
        return retval

    __enter__ = acquire

    def release(self) -> None:
        """Releases the lock."""
        caller = sys._getframe().f_back
        if caller is not None:
            caller2 = caller.f_code.co_name
        else:
            caller2 = "*Unknown*"
        thread_name = current_thread().name
        LOG.debug(
            "attempting to release lock %s from %s in thread %s",
            self._lock_name,
            caller2,
            thread_name,
        )
        self._Rlock.release()
        LOG.debug(
            "released lock %s from %s in thread %s info: %s",
            self._lock_name,
            caller2,
            thread_name,
            repr(self._Rlock),
        )

    def __exit__(self, t, v, b):
        """Automatically release lock on exit.

        Args:
            t (_type_): _description_
            v (_type_): _description_
            b (_type_): _description_
        """
        caller = sys._getframe().f_back
        if caller is not None:
            caller2 = caller.f_code.co_name
        else:
            caller2 = "*Unknown*"
        thread_name = current_thread().name
        LOG.debug(
            "released lock %s from %s in thread %s at exit",
            self._lock_name,
            caller2,
            thread_name,
        )

        self._Rlock.release()


def parse_pulse_datetime(datestring: str) -> datetime:
    """Parse pulse date strings.

    Args:
        datestring (str): the string to parse

    Raises:
        ValueError: pass through of value error if string
                    cannot be converted

    Returns:
        datetime: time value of given string
    """
    datestring = datestring.replace("\xa0", " ").rstrip()
    split_string = [s for s in datestring.split(" ") if s.strip()]
    if len(split_string) < 3:
        raise ValueError("Invalid datestring")
    t = datetime.today()
    if split_string[0].lstrip() == "Today":
        last_update = t
    elif split_string[0].lstrip() == "Yesterday":
        last_update = t - timedelta(days=1)
    else:
        tempdate = f"{split_string[0]}/{t.year}"
        last_update = datetime.strptime(tempdate, "%m/%d/%Y")
    if last_update > t:
        last_update = last_update.replace(year=t.year)
    update_time = datetime.time(
        datetime.strptime(split_string[1] + split_string[2], "%I:%M%p")
    )
    last_update = datetime.combine(last_update, update_time)
    return last_update


def set_debug_lock(debug_lock: bool, name: str) -> "RLock | DebugRLock":
    """Set lock or debug lock

    Args:
        debug_lock (bool): set a debug lock
        name (str): debug lock name

    Returns:
        RLock | DebugRLock: lock object to return
    """
    if debug_lock:
        return DebugRLock(name)
    return RLock()



"""Base Python Class for pyadtpulse."""

import logging
import asyncio
import time
from threading import RLock, Thread
from warnings import warn

import uvloop

from .const import (
    ADT_DEFAULT_HTTP_USER_AGENT,
    ADT_DEFAULT_KEEPALIVE_INTERVAL,
    ADT_DEFAULT_RELOGIN_INTERVAL,
    DEFAULT_API_HOST,
)
from .pyadtpulse_async import SYNC_CHECK_TASK_NAME, PyADTPulseAsync
from .util import DebugRLock, set_debug_lock

LOG = logging.getLogger(__name__)


class PyADTPulse(PyADTPulseAsync):
    """Base object for ADT Pulse service."""

    __slots__ = ("_session_thread", "_p_attribute_lock", "_login_exception")

    def __init__(
        self,
        username: str,
        password: str,
        fingerprint: str,
        service_host: str = DEFAULT_API_HOST,
        user_agent=ADT_DEFAULT_HTTP_USER_AGENT["User-Agent"],
        do_login: bool = True,
        debug_locks: bool = False,
        keepalive_interval: int = ADT_DEFAULT_KEEPALIVE_INTERVAL,
        relogin_interval: int = ADT_DEFAULT_RELOGIN_INTERVAL,
        detailed_debug_logging: bool = False,
    ):
        self._p_attribute_lock = set_debug_lock(
            debug_locks, "pyadtpulse._p_attribute_lockattribute_lock"
        )
        warn(
            "PyADTPulse is deprecated, please use PyADTPulseAsync instead",
            DeprecationWarning,
            stacklevel=2,
        )
        super().__init__(
            username,
            password,
            fingerprint,
            service_host,
            user_agent,
            debug_locks,
            keepalive_interval,
            relogin_interval,
            detailed_debug_logging,
        )
        self._session_thread: Thread | None = None
        self._login_exception: Exception | None = None
        if do_login:
            self.login()

    def __repr__(self) -> str:
        """Object representation."""
        return (
            f"<{self.__class__.__name__}: {self._authentication_properties.username}>"
        )

    # ADTPulse API endpoint is configurable (besides default US ADT Pulse endpoint) to
    # support testing as well as alternative ADT Pulse endpoints such as
    # portal-ca.adtpulse.com

    def _pulse_session_thread(self) -> None:
        """
        Pulse the session thread.

        Acquires the attribute lock and creates a background thread for the ADT
        Pulse API. The thread runs the synchronous loop `_sync_loop()` until completion.
        Once the loop finishes, the thread is closed, the pulse connection's event loop
        is set to `None`, and the session thread is set to `None`.
        """
        # lock is released in sync_loop()
        self._p_attribute_lock.acquire()

        LOG.debug("Creating ADT Pulse background thread")
        asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())
        loop = asyncio.new_event_loop()
        self._pulse_connection_properties.loop = loop
        loop.run_until_complete(self._sync_loop())

        loop.close()
        self._pulse_connection_properties.loop = None
        self._session_thread = None

    async def _sync_loop(self) -> None:
        """
        Asynchronous function that represents the main loop of the synchronization
        process.

        This function is responsible for executing the synchronization logic. It starts
        by calling the `async_login` method to perform the login operation. After that,
        it releases the `_p_attribute_lock` to allow other tasks to access the
        attributes.
        If the login operation was successful, it waits for the `_timeout_task` to
        complete using the `asyncio.wait` function.  If the `_timeout_task` is not set,
        it raises a `RuntimeError` to indicate that background tasks were not created.

        After the waiting process, it enters a while loop that continues as long as the
        `_authenticated` event is set. Inside the loop, it waits for 0.5 seconds using
        the `asyncio.sleep` function. This wait allows the logout process to complete
        before continuing with the synchronization logic.
        """
        try:
            await self.async_login()
        except Exception as e:
            self._login_exception = e
        self._p_attribute_lock.release()
        if self._login_exception is not None:
            return
        if self._timeout_task is not None:
            task_list = (self._timeout_task,)
            try:
                await asyncio.wait(task_list)
            except asyncio.CancelledError:
                pass
            except Exception as e:  # pylint: disable=broad-except
                LOG.exception(
                    "Received exception while waiting for ADT Pulse service %s", e
                )
        else:
            # we should never get here
            raise RuntimeError("Background pyadtpulse tasks not created")
        while self._pulse_connection_status.authenticated_flag.is_set():
            # busy wait until logout is done
            await asyncio.sleep(0.5)

    def login(self) -> None:
        """Login to ADT Pulse and generate access token.

        Raises:
            Exception from async_login
        """
        self._p_attribute_lock.acquire()
        # probably shouldn't be a daemon thread
        self._session_thread = thread = Thread(
            target=self._pulse_session_thread,
            name="PyADTPulse Session",
            daemon=True,
        )
        self._p_attribute_lock.release()

        self._session_thread.start()
        time.sleep(1)

        # thread will unlock after async_login, so attempt to obtain
        # lock to block current thread until then
        # if it's still alive, no exception
        self._p_attribute_lock.acquire()
        self._p_attribute_lock.release()
        if not thread.is_alive():
            if self._login_exception is not None:
                raise self._login_exception

    def logout(self) -> None:
        """Log out of ADT Pulse."""
        loop = self._pulse_connection.check_sync(
            "Attempting to call sync logout without sync login"
        )
        sync_thread = self._session_thread

        coro = self.async_logout()
        asyncio.run_coroutine_threadsafe(coro, loop)
        if sync_thread is not None:
            sync_thread.join()

    @property
    def attribute_lock(self) -> "RLock| DebugRLock":
        """Get attribute lock for PyADTPulse object.

        Returns:
            RLock: thread Rlock
        """
        return self._p_attribute_lock

    @property
    def loop(self) -> asyncio.AbstractEventLoop | None:
        """Get event loop.

        Returns:
            Optional[asyncio.AbstractEventLoop]: the event loop object or
                                                 None if no thread is running
        """
        return self._pulse_connection_properties.loop

    @property
    def updates_exist(self) -> bool:
        """Check if updated data exists.

        Returns:
            bool: True if updated data exists
        """
        with self._p_attribute_lock:
            if self._sync_task is None:
                loop = self._pulse_connection_properties.loop
                if loop is None:
                    raise RuntimeError(
                        "ADT pulse sync function updates_exist() "
                        "called from async session"
                    )
                coro = self._sync_check_task()
                self._sync_task = loop.create_task(
                    coro, name=f"{SYNC_CHECK_TASK_NAME}: Sync session"
                )
            if self._pulse_properties.updates_exist.is_set():
                self._pulse_properties.updates_exist.clear()
                return True
            return False

    def update(self) -> bool:
        """Update ADT Pulse data.

        Returns:
            bool: True on success
        """
        coro = self.async_update()
        return asyncio.run_coroutine_threadsafe(
            coro,
            self._pulse_connection.check_sync(
                "Attempting to run sync update from async login"
            ),
        ).result()

    async def async_login(self) -> None:
        self._pulse_connection_properties.check_async(
            "Cannot login asynchronously with a synchronous session"
        )
        await super().async_login()

    async def async_logout(self) -> None:
        self._pulse_connection_properties.check_async(
            "Cannot logout asynchronously with a synchronous session"
        )
        await super().async_logout()

    async def async_update(self) -> bool:
        self._pulse_connection_properties.check_async(
            "Cannot update asynchronously with a synchronous session"
        )
        return await super().async_update()



"""PyADTPulse Properties."""

import logging
import asyncio
from warnings import warn

from typeguard import typechecked

from .const import (
    ADT_DEFAULT_KEEPALIVE_INTERVAL,
    ADT_DEFAULT_RELOGIN_INTERVAL,
    ADT_MAX_KEEPALIVE_INTERVAL,
    ADT_MIN_RELOGIN_INTERVAL,
)
from .site import ADTPulseSite
from .util import set_debug_lock

LOG = logging.getLogger(__name__)


class PyADTPulseProperties:
    """PyADTPulse Properties."""

    __slots__ = (
        "_updates_exist",
        "_pp_attribute_lock",
        "_relogin_interval",
        "_keepalive_interval",
        "_site",
    )

    @staticmethod
    @typechecked
    def _check_keepalive_interval(keepalive_interval: int) -> None:
        if keepalive_interval > ADT_MAX_KEEPALIVE_INTERVAL or keepalive_interval <= 0:
            raise ValueError(
                f"keepalive interval ({keepalive_interval}) must be "
                f"greater than 0 and less than {ADT_MAX_KEEPALIVE_INTERVAL}"
            )

    @staticmethod
    @typechecked
    def _check_relogin_interval(relogin_interval: int) -> None:
        if relogin_interval < ADT_MIN_RELOGIN_INTERVAL:
            raise ValueError(
                f"relogin interval ({relogin_interval}) must be "
                f"greater than {ADT_MIN_RELOGIN_INTERVAL}"
            )

    @typechecked
    def __init__(
        self,
        keepalive_interval: int = ADT_DEFAULT_KEEPALIVE_INTERVAL,
        relogin_interval: int = ADT_DEFAULT_RELOGIN_INTERVAL,
        debug_locks: bool = False,
    ) -> None:
        """Create a PyADTPulse properties object.
        Args:
        pulse_authentication_properties (PulseAuthenticationProperties):
            an instance of PulseAuthenticationProperties
        pulse_connection_properties (PulseConnectionProperties):
        """
        # FIXME use thread event/condition, regular condition?
        # defer initialization to make sure we have an event loop

        self._updates_exist = asyncio.locks.Event()

        self._pp_attribute_lock = set_debug_lock(
            debug_locks, "pyadtpulse.async_attribute_lock"
        )

        self._site: ADTPulseSite | None = None
        self.keepalive_interval = keepalive_interval
        self.relogin_interval = relogin_interval

    @property
    def relogin_interval(self) -> int:
        """Get re-login interval.

        Returns:
            int: number of minutes to re-login to Pulse
                 0 means disabled
        """
        with self._pp_attribute_lock:
            return self._relogin_interval

    @relogin_interval.setter
    @typechecked
    def relogin_interval(self, interval: int | None) -> None:
        """Set re-login interval.

        Args:
            interval (int|None): The number of minutes between logins.
                            If set to None, resets to ADT_DEFAULT_RELOGIN_INTERVAL

        Raises:
            ValueError: if a relogin interval of less than ADT_MIN_RELOGIN_INTERVAL
                        minutes is specified
        """
        if interval is None:
            interval = ADT_DEFAULT_RELOGIN_INTERVAL
        else:
            self._check_relogin_interval(interval)
        with self._pp_attribute_lock:
            self._relogin_interval = interval
            LOG.debug("relogin interval set to %d", self._relogin_interval)

    @property
    def keepalive_interval(self) -> int:
        """Get the keepalive interval in minutes.

        Returns:
            int: the keepalive interval
        """
        with self._pp_attribute_lock:
            return self._keepalive_interval

    @keepalive_interval.setter
    @typechecked
    def keepalive_interval(self, interval: int | None) -> None:
        """Set the keepalive interval in minutes.

        Args:
            interval (int|None): The number of minutes between keepalive calls
                                 If set to None, resets to ADT_DEFAULT_KEEPALIVE_INTERVAL

        Raises:
            ValueError: if a keepalive interval of greater than ADT_MAX_KEEPALIVE_INTERVAL
                        minutes is specified
        """
        if interval is None:
            interval = ADT_DEFAULT_KEEPALIVE_INTERVAL
        else:
            self._check_keepalive_interval(interval)
        with self._pp_attribute_lock:
            self._keepalive_interval = interval
            LOG.debug("keepalive interval set to %d", self._keepalive_interval)

    @property
    def sites(self) -> list[ADTPulseSite]:
        """Return all sites for this ADT Pulse account."""
        warn(
            "multiple sites being removed, use pyADTPulse.site instead",
            PendingDeprecationWarning,
            stacklevel=2,
        )
        with self._pp_attribute_lock:
            if self._site is None:
                raise RuntimeError(
                    "No sites have been retrieved, have you logged in yet?"
                )
            return [self._site]

    @property
    def site(self) -> ADTPulseSite:
        """Return the site associated with the Pulse login."""
        with self._pp_attribute_lock:
            if self._site is None:
                raise RuntimeError(
                    "No sites have been retrieved, have you logged in yet?"
                )
            return self._site

    def set_update_status(self) -> None:
        """Sets updates_exist to notify wait_for_update."""
        with self._pp_attribute_lock:
            self.updates_exist.set()

    @property
    def updates_exist(self) -> asyncio.locks.Event:
        """Check if updates exist."""
        with self._pp_attribute_lock:
            return self._updates_exist



"""Pulse Connection Status."""

from asyncio import Event

from typeguard import typechecked

from .pulse_backoff import PulseBackoff
from .util import set_debug_lock


class PulseConnectionStatus:
    """Pulse Connection Status."""

    __slots__ = (
        "_backoff",
        "_authenticated_flag",
        "_pcs_attribute_lock",
    )

    @typechecked
    def __init__(self, debug_locks: bool = False, detailed_debug_logging=False):
        self._pcs_attribute_lock = set_debug_lock(
            debug_locks, "pyadtpulse.pcs_attribute_lock"
        )
        """Initialize the connection status object.

        Args:
            debug_locks (bool, optional): Enable debug locks. Defaults to False.
            detailed_debug_logging (bool, optional): Enable detailed debug logging for the backoff.
                Defaults to False.
        """
        self._backoff = PulseBackoff(
            "Connection Status",
            initial_backoff_interval=1,
            detailed_debug_logging=detailed_debug_logging,
        )
        self._authenticated_flag = Event()

    @property
    def authenticated_flag(self) -> Event:
        """Get the authenticated flag."""
        with self._pcs_attribute_lock:
            return self._authenticated_flag

    @property
    def retry_after(self) -> float:
        """Get the number of seconds to wait before retrying HTTP requests."""
        with self._pcs_attribute_lock:
            return self._backoff.expiration_time

    @retry_after.setter
    @typechecked
    def retry_after(self, seconds: float) -> None:
        """Set time after which HTTP requests can be retried."""
        with self._pcs_attribute_lock:
            self._backoff.set_absolute_backoff_time(seconds)

    def get_backoff(self) -> PulseBackoff:
        """Get the backoff object."""
        return self._backoff

    @property
    def detailed_debug_logging(self) -> bool:
        """Get the detailed debug logging flag."""
        with self._pcs_attribute_lock:
            return self._backoff.detailed_debug_logging

    @detailed_debug_logging.setter
    @typechecked
    def detailed_debug_logging(self, value: bool):
        """Set the detailed debug logging flag."""
        with self._pcs_attribute_lock:
            self._backoff.detailed_debug_logging = value



"""Pulse connection info."""

from asyncio import AbstractEventLoop
from re import search

from aiohttp import ClientSession
from typeguard import typechecked

from .const import (
    ADT_DEFAULT_HTTP_ACCEPT_HEADERS,
    ADT_DEFAULT_HTTP_USER_AGENT,
    ADT_DEFAULT_SEC_FETCH_HEADERS,
    API_HOST_CA,
    API_PREFIX,
    DEFAULT_API_HOST,
)
from .util import set_debug_lock


class PulseConnectionProperties:
    """Pulse connection info."""

    __slots__ = (
        "_api_host",
        "_session",
        "_user_agent",
        "_loop",
        "_api_version",
        "_pci_attribute_lock",
        "_detailed_debug_logging",
        "_debug_locks",
    )

    @staticmethod
    @typechecked
    def check_service_host(service_host: str) -> None:
        """Check if service host is valid."""
        if service_host is None or service_host == "":
            raise ValueError("Service host is mandatory")
        if service_host not in (DEFAULT_API_HOST, API_HOST_CA):
            raise ValueError(
                f"Service host must be one of {DEFAULT_API_HOST}" f" or {API_HOST_CA}"
            )

    @staticmethod
    def get_api_version(response_path: str) -> str | None:
        """Regex used to exctract the API version.

        Use for testing.
        """
        version: str | None = None
        if not response_path:
            return None
        m = search(f"{API_PREFIX}(.+)/[a-z]*/", response_path)
        if m is not None:
            version = m.group(1)
        return version

    def __init__(
        self,
        host: str,
        user_agent=ADT_DEFAULT_HTTP_USER_AGENT["User-Agent"],
        detailed_debug_logging=False,
        debug_locks=False,
    ) -> None:
        """Initialize Pulse connection information."""
        self._pci_attribute_lock = set_debug_lock(
            debug_locks, "pyadtpulse.pci_attribute_lock"
        )
        self.debug_locks = debug_locks
        self.detailed_debug_logging = detailed_debug_logging
        self._loop: AbstractEventLoop | None = None
        self._session: ClientSession | None = None
        self.service_host = host
        self._api_version = ""
        self._user_agent = user_agent

    def __del__(self):
        """Destructor for ADTPulseConnection."""
        if self._session is not None and not self._session.closed:
            self._session.detach()

    def _set_headers(self) -> None:
        if self._session is not None:
            self._session.headers.update(ADT_DEFAULT_HTTP_ACCEPT_HEADERS)
            self._session.headers.update(ADT_DEFAULT_SEC_FETCH_HEADERS)
            self._session.headers.update({"User-Agent": self._user_agent})

    @property
    def service_host(self) -> str:
        """Get the service host."""
        with self._pci_attribute_lock:
            return self._api_host

    @service_host.setter
    @typechecked
    def service_host(self, host: str):
        """Set the service host.

        Raises:
            ValueError if host is not valid.
        """
        self.check_service_host(host)
        with self._pci_attribute_lock:
            self._api_host = host

    @property
    def detailed_debug_logging(self) -> bool:
        """Get the detailed debug logging flag."""
        with self._pci_attribute_lock:
            return self._detailed_debug_logging

    @detailed_debug_logging.setter
    @typechecked
    def detailed_debug_logging(self, value: bool):
        """Set the detailed debug logging flag."""
        with self._pci_attribute_lock:
            self._detailed_debug_logging = value

    @property
    def debug_locks(self) -> bool:
        """Get the debug locks flag."""
        with self._pci_attribute_lock:
            return self._debug_locks

    @debug_locks.setter
    @typechecked
    def debug_locks(self, value: bool):
        """Set the debug locks flag."""
        with self._pci_attribute_lock:
            self._debug_locks = value

    @typechecked
    def check_sync(self, message: str) -> AbstractEventLoop:
        """Checks if sync login was performed.

        Returns the loop to use for run_coroutine_threadsafe if so.
        Raises RuntimeError with given message if not.
        """
        with self._pci_attribute_lock:
            if self._loop is None:
                raise RuntimeError(message)
            return self._loop

    @typechecked
    def check_async(self, message: str) -> None:
        """Checks if async login was performed.

        Raises RuntimeError with given message if not.
        """
        with self._pci_attribute_lock:
            if self._loop is not None:
                raise RuntimeError(message)

    @property
    def loop(self) -> AbstractEventLoop | None:
        """Get the event loop."""
        with self._pci_attribute_lock:
            return self._loop

    @loop.setter
    @typechecked
    def loop(self, loop: AbstractEventLoop | None):
        """Set the event loop."""
        with self._pci_attribute_lock:
            self._loop = loop

    @property
    def session(self) -> ClientSession:
        """Get the session."""
        with self._pci_attribute_lock:
            if self._session is None:
                self._session = ClientSession()
            self._set_headers()
            return self._session

    @property
    def api_version(self) -> str:
        """Get the API version."""
        with self._pci_attribute_lock:
            return self._api_version

    @api_version.setter
    @typechecked
    def api_version(self, version: str):
        """Set the API version.

        Raises:
            ValueError: if version is not in the form major.minor.patch-subpatch
        """

        def check_version_string(value: str):
            parts = value.split("-")
            if len(parts) == 2:
                version_parts = parts[0].split(".")
                if not (
                    version_parts[0].isdigit()
                    and version_parts[1].isdigit()
                    and version_parts[2].isdigit()
                    and parts[1].isdigit()
                ):
                    raise ValueError(
                        "API version must be in the form major.minor.patch-subpatch"
                    )
                if len(version_parts) == 3 and version_parts[0].isdigit():
                    major_version = int(version_parts[0])
                    if major_version >= 26:
                        return
                    else:
                        raise ValueError("API version is numeric but less than 26")
            raise ValueError(
                "API version must be in the form major.minor.patch-subpatch"
            )

        with self._pci_attribute_lock:
            check_version_string(version)
            self._api_version = version

    @typechecked
    def make_url(self, uri: str) -> str:
        """Create a URL to service host from a URI.

        Args:
            uri (str): the URI to convert

        Returns:
            str: the converted string
        """
        with self._pci_attribute_lock:
            return f"{self._api_host}{API_PREFIX}{self._api_version}{uri}"

    async def clear_session(self):
        """Clear the session."""
        with self._pci_attribute_lock:
            old_session = self._session
            self._session = None
            if old_session:
                await old_session.close()



"""Pulse Site Properties."""

from threading import RLock
from warnings import warn

from typeguard import typechecked

from .alarm_panel import ADTPulseAlarmPanel
from .gateway import ADTPulseGateway
from .util import DebugRLock, set_debug_lock
from .zones import ADTPulseFlattendZone, ADTPulseZones


class ADTPulseSiteProperties:
    """Pulse Site Properties."""

    __slots__ = (
        "_id",
        "_name",
        "_last_updated",
        "_alarm_panel",
        "_zones",
        "_site_lock",
        "_gateway",
    )

    @typechecked
    def __init__(self, site_id: str, name: str, debug_locks: bool = False):
        self._id = site_id
        self._name = name
        self._last_updated: int = 0
        self._zones = ADTPulseZones()
        self._site_lock: RLock | DebugRLock
        self._site_lock = set_debug_lock(debug_locks, "pyadtpulse.site_property_lock")
        self._alarm_panel = ADTPulseAlarmPanel()
        self._gateway = ADTPulseGateway()

    @property
    def id(self) -> str:
        """Get site id.

        Returns:
            str: the site id
        """
        return self._id

    @property
    def name(self) -> str:
        """Get site name.

        Returns:
            str: the site name
        """
        return self._name

    # FIXME: should this actually return if the alarm is going off!?  How do we
    # return state that shows the site is compromised??

    @property
    def last_updated(self) -> int:
        """Return time site last updated.

        Returns:
            int: the time site last updated as datetime
        """
        with self._site_lock:
            return self._last_updated

    @property
    def site_lock(self) -> "RLock| DebugRLock":
        """Get thread lock for site data.

        Not needed for async

        Returns:
            RLock: thread RLock
        """
        return self._site_lock

    @property
    def zones(self) -> list[ADTPulseFlattendZone] | None:
        """Return all zones registered with the ADT Pulse account.

        (cached copy of last fetch)
        See Also fetch_zones()
        """
        with self._site_lock:
            if not self._zones:
                raise RuntimeError("No zones exist")
            return self._zones.flatten()

    @property
    def zones_as_dict(self) -> ADTPulseZones | None:
        """Return zone information in dictionary form.

        Returns:
            ADTPulseZones: all zone information
        """
        with self._site_lock:
            if not self._zones:
                raise RuntimeError("No zones exist")
            return self._zones

    @property
    def alarm_control_panel(self) -> ADTPulseAlarmPanel:
        """Return the alarm panel object for the site.

        Returns:
            Optional[ADTPulseAlarmPanel]: the alarm panel object
        """
        return self._alarm_panel

    @property
    def gateway(self) -> ADTPulseGateway:
        """Get gateway device object.

        Returns:
            ADTPulseGateway: Gateway device
        """
        return self._gateway

    @property
    def updates_may_exist(self) -> bool:
        """Query whether updated sensor data exists.

        Deprecated, use method on pyADTPulse object instead
        """
        # FIXME: this should actually capture the latest version
        # and compare if different!!!
        # ...this doesn't actually work if other components are also checking
        #  if updates exist
        warn(
            "updates_may_exist on site object is deprecated, "
            "use method on pyADTPulse object instead",
            DeprecationWarning,
            stacklevel=2,
        )
        return False

    async def async_update(self) -> bool:
        """Force update site/zone data async with current data.

        Deprecated, use method on pyADTPulse object instead
        """
        warn(
            "updating zones from site object is deprecated, "
            "use method on pyADTPulse object instead",
            DeprecationWarning,
            stacklevel=2,
        )
        return False

    def update(self) -> bool:
        """Force update site/zones with current data.

        Deprecated, use method on pyADTPulse object instead
        """
        warn(
            "updating zones from site object is deprecated, "
            "use method on pyADTPulse object instead",
            DeprecationWarning,
            stacklevel=2,
        )
        return False



"""ADT Pulse Async API."""

import logging
import asyncio
import re
import time
from random import randint
from warnings import warn

from bs4 import BeautifulSoup
from typeguard import typechecked
from yarl import URL

from .alarm_panel import ADT_ALARM_UNKNOWN
from .const import (
    ADT_DEFAULT_HTTP_USER_AGENT,
    ADT_DEFAULT_KEEPALIVE_INTERVAL,
    ADT_DEFAULT_RELOGIN_INTERVAL,
    ADT_GATEWAY_STRING,
    ADT_SYNC_CHECK_URI,
    ADT_TIMEOUT_URI,
    DEFAULT_API_HOST,
)
from .exceptions import (
    PulseAccountLockedError,
    PulseAuthenticationError,
    PulseClientConnectionError,
    PulseGatewayOfflineError,
    PulseMFARequiredError,
    PulseNotLoggedInError,
    PulseServerConnectionError,
    PulseServiceTemporarilyUnavailableError,
)
from .pulse_authentication_properties import PulseAuthenticationProperties
from .pulse_connection import PulseConnection
from .pulse_connection_properties import PulseConnectionProperties
from .pulse_connection_status import PulseConnectionStatus
from .pyadtpulse_properties import PyADTPulseProperties
from .site import ADTPulseSite
from .util import handle_response, set_debug_lock

LOG = logging.getLogger(__name__)
SYNC_CHECK_TASK_NAME = "ADT Pulse Sync Check Task"
KEEPALIVE_TASK_NAME = "ADT Pulse Keepalive Task"
# backoff time before warning in wait_for_update()
WARN_TRANSIENT_FAILURE_THRESHOLD = 2


class PyADTPulseAsync:
    """ADT Pulse Async API."""

    __slots__ = (
        "_sync_task",
        "_timeout_task",
        "_pa_attribute_lock",
        "_pulse_properties",
        "_authentication_properties",
        "_pulse_connection_properties",
        "_pulse_connection",
        "_pulse_connection_status",
        "_site",
        "_detailed_debug_logging",
        "_sync_check_exception",
        "_sync_check_sleeping",
    )

    @typechecked
    def __init__(
        self,
        username: str,
        password: str,
        fingerprint: str,
        service_host: str = DEFAULT_API_HOST,
        user_agent=ADT_DEFAULT_HTTP_USER_AGENT["User-Agent"],
        debug_locks: bool = False,
        keepalive_interval: int = ADT_DEFAULT_KEEPALIVE_INTERVAL,
        relogin_interval: int = ADT_DEFAULT_RELOGIN_INTERVAL,
        detailed_debug_logging: bool = False,
    ) -> None:
        """Create a PyADTPulse object.
        Args:
            username (str): Username.
            password (str): Password.
            fingerprint (str): 2FA fingerprint.
            service_host (str, optional): host prefix to use
                         i.e. https://portal.adtpulse.com or
                              https://portal-ca.adtpulse.com
            user_agent (str, optional): User Agent.
                         Defaults to ADT_DEFAULT_HTTP_HEADERS["User-Agent"].
            debug_locks: (bool, optional): use debugging locks
                        Defaults to False
            keepalive_interval (int, optional): number of minutes between
                        keepalive checks, defaults to ADT_DEFAULT_KEEPALIVE_INTERVAL,
                        maxiumum is ADT_MAX_KEEPALIVE_INTERVAL
            relogin_interval (int, optional): number of minutes between relogin checks
                        defaults to ADT_DEFAULT_RELOGIN_INTERVAL,
                        minimum is ADT_MIN_RELOGIN_INTERVAL
            detailed_debug_logging (bool, optional): enable detailed debug logging
        """
        self._pa_attribute_lock = set_debug_lock(
            debug_locks, "pyadtpulse.pa_attribute_lock"
        )
        self._pulse_connection_properties = PulseConnectionProperties(
            service_host, user_agent, detailed_debug_logging, debug_locks
        )
        self._authentication_properties = PulseAuthenticationProperties(
            username=username,
            password=password,
            fingerprint=fingerprint,
            debug_locks=debug_locks,
        )
        self._pulse_connection_status = PulseConnectionStatus(
            debug_locks=debug_locks, detailed_debug_logging=detailed_debug_logging
        )
        self._pulse_properties = PyADTPulseProperties(
            keepalive_interval=keepalive_interval,
            relogin_interval=relogin_interval,
            debug_locks=debug_locks,
        )
        self._pulse_connection = PulseConnection(
            self._pulse_connection_status,
            self._pulse_connection_properties,
            self._authentication_properties,
            debug_locks,
        )
        self._sync_task: asyncio.Task | None = None
        self._timeout_task: asyncio.Task | None = None
        self._site: ADTPulseSite | None = None
        self._detailed_debug_logging = detailed_debug_logging
        pc_backoff = self._pulse_connection.get_login_backoff()
        self._sync_check_exception: Exception | None = PulseNotLoggedInError()
        pc_backoff.reset_backoff()
        self._sync_check_sleeping = asyncio.Event()

    def __repr__(self) -> str:
        """Object representation."""
        return (
            f"<{self.__class__.__name__}: {self._authentication_properties.username}>"
        )

    async def _update_sites(self, soup: BeautifulSoup) -> None:
        with self._pa_attribute_lock:
            start_time = 0.0
            if self._pulse_connection.detailed_debug_logging:
                start_time = time.time()
            if self._site is None:
                await self._initialize_sites(soup)
                if self._site is None:
                    raise RuntimeError("pyadtpulse could not retrieve site")
            self._site.alarm_control_panel.update_alarm_from_soup(soup)
            self._site.update_zone_from_soup(soup)
            if self._pulse_connection.detailed_debug_logging:
                LOG.debug(
                    "Updated site %s in %s seconds",
                    self._site.id,
                    time.time() - start_time,
                )

    async def _initialize_sites(self, soup: BeautifulSoup) -> None:
        """
        Initializes the sites in the ADT Pulse account.

        Args:
            soup (BeautifulSoup): The parsed HTML soup object.

        Raises:
            PulseGatewayOfflineError: if the gateway is offline
        """
        # typically, ADT Pulse accounts have only a single site (premise/location)
        single_premise = soup.find("span", {"id": "p_singlePremise"})
        if single_premise:
            site_name = single_premise.text
            start_time = 0.0
            if self._pulse_connection.detailed_debug_logging:
                start_time = time.time()
            # FIXME: this code works, but it doesn't pass the linter
            signout_link = str(
                soup.find("a", {"class": "p_signoutlink"}).get("href")  # type: ignore
            )
            if signout_link:
                m = re.search("networkid=(.+)&", signout_link)
                if m and m.group(1) and m.group(1):
                    site_id = m.group(1)
                    LOG.debug("Discovered site id %s: %s", site_id, site_name)
                    new_site = ADTPulseSite(self._pulse_connection, site_id, site_name)

                    # fetch zones first, so that we can have the status
                    # updated with _update_alarm_status
                    if not await new_site.fetch_devices(None):
                        LOG.error("Could not fetch zones from ADT site")
                    new_site.alarm_control_panel.update_alarm_from_soup(soup)
                    if new_site.alarm_control_panel.status == ADT_ALARM_UNKNOWN:
                        new_site.gateway.is_online = False
                    new_site.update_zone_from_soup(soup)
                    self._site = new_site
                    if self._pulse_connection.detailed_debug_logging:
                        LOG.debug(
                            "Initialized site %s in %s seconds",
                            self._site.id,
                            time.time() - start_time,
                        )
                    return
            else:
                LOG.warning(
                    "Couldn't find site id for %s in %s", site_name, signout_link
                )
        else:
            LOG.error("ADT Pulse accounts with MULTIPLE sites not supported!!!")

    # ...and current network id from:
    # <a id="p_signout1" class="p_signoutlink"
    # href="/myhome/16.0.0-131/access/signout.jsp?networkid=150616za043597&partner=adt"
    # onclick="return flagSignOutInProcess();">
    #
    # ... or perhaps better, just extract all from /system/settings.jsp

    def _get_task_name(self, task: asyncio.Task | None, default_name) -> str:
        """
        Get the name of a task.

        Parameters:
            task (Task): The task object.
            default_name (str): The default name to use if the task is None.

        Returns:
            str: The name of the task if it is not None, otherwise the default name
            with a suffix indicating a possible internal error.
        """
        if task is not None:
            return task.get_name()
        return f"{default_name} - possible internal error"

    def _get_sync_task_name(self) -> str:
        return self._get_task_name(self._sync_task, SYNC_CHECK_TASK_NAME)

    def _get_timeout_task_name(self) -> str:
        return self._get_task_name(self._timeout_task, KEEPALIVE_TASK_NAME)

    def _set_update_exception(self, e: Exception | None) -> None:
        self.sync_check_exception = e
        self._pulse_properties.updates_exist.set()

    async def _keepalive_task(self) -> None:
        """
        Asynchronous function that runs a keepalive task to maintain the connection
        with the ADT Pulse cloud.
        """

        async def reset_pulse_cloud_timeout() -> tuple[int, str | None, URL | None]:
            return await self._pulse_connection.async_query(ADT_TIMEOUT_URI, "POST")

        async def update_gateway_device_if_needed() -> None:
            if self.site.gateway.next_update < time.time():
                await self.site.set_device(ADT_GATEWAY_STRING)

        def should_relogin(relogin_interval: int) -> bool:
            return (
                relogin_interval != 0
                and time.time() - self._authentication_properties.last_login_time
                > randint(int(0.75 * relogin_interval), relogin_interval)
            )

        next_full_logout_time = time.time() + 24 * 60 * 60
        response: str | None
        task_name: str = self._get_task_name(self._timeout_task, KEEPALIVE_TASK_NAME)
        LOG.debug("creating %s", task_name)

        while True:
            relogin_interval = self._pulse_properties.relogin_interval * 60
            try:
                await asyncio.sleep(self._pulse_properties.keepalive_interval * 60)
                if self._pulse_connection_status.retry_after > time.time():
                    LOG.debug(
                        "%s: Skipping actions because retry_after > now", task_name
                    )
                    continue
                if not self._pulse_connection.is_connected:
                    LOG.debug("%s: Skipping relogin because not connected", task_name)
                    continue
                if should_relogin(relogin_interval):
                    msg = "quick"
                    if time.time() > next_full_logout_time:
                        msg = "full"
                    with self._pa_attribute_lock:
                        if self._sync_task:
                            if self._detailed_debug_logging:
                                LOG.debug(
                                    "%s: waiting for sync check task to sleep",
                                    task_name,
                                )
                            await self._sync_check_sleeping.wait()
                    if msg == "full":
                        next_full_logout_time = time.time() + 24 * 60 * 60
                        await self.async_logout()
                    else:
                        await self._pulse_connection.quick_logout()
                    LOG.debug("%s: performing %s logout", task_name, msg)
                    try:
                        await self._login_looped(task_name)
                    except (PulseAuthenticationError, PulseMFARequiredError) as ex:
                        LOG.error("%s task exiting due to %s", task_name, ex.args[0])
                        return
                    continue
                LOG.debug("Resetting timeout")
                try:
                    code, response, url = await reset_pulse_cloud_timeout()
                except (
                    PulseServiceTemporarilyUnavailableError,
                    PulseClientConnectionError,
                    PulseServerConnectionError,
                ) as ex:
                    LOG.debug(
                        "Could not reset ADT Pulse cloud timeout due to %s, skipping",
                        ex.args[0],
                    )
                    continue
                if (
                    not handle_response(
                        code,
                        url,
                        logging.WARNING,
                        "Could not reset ADT Pulse cloud timeout",
                    )
                    or response is None
                ):
                    continue
                await update_gateway_device_if_needed()

            except asyncio.CancelledError:
                LOG.debug("%s cancelled", task_name)
                return

    async def _clean_done_tasks(self) -> None:
        with self._pa_attribute_lock:
            if self._sync_task is not None and self._sync_task.done():
                await self._sync_task
                self._sync_task = None
            if self._timeout_task is not None and self._timeout_task.done():
                await self._timeout_task
                self._timeout_task = None

    async def _cancel_task(self, task: asyncio.Task | None) -> None:
        """
        Cancel a given asyncio task.

        Args:
            task (asyncio.Task | None): The task to be cancelled.
        """
        await self._clean_done_tasks()
        if task is None:
            return
        task_name = task.get_name()
        LOG.debug("cancelling %s", task_name)
        task.cancel()
        try:
            await task
        except asyncio.CancelledError:
            pass
        if task == self._sync_task:
            with self._pa_attribute_lock:
                self._sync_task = None
        else:
            with self._pa_attribute_lock:
                self._timeout_task = None
        LOG.debug("%s successfully cancelled", task_name)

    async def _login_looped(self, task_name: str) -> None:
        """
        Logs in and loops until successful.

        Args:
            None.
        Returns:
            None
        """
        count = 0
        log_level = logging.DEBUG

        while True:
            count += 1
            if count > 5:
                log_level = logging.WARNING
            LOG.log(log_level, "%s performming loop login", task_name)
            try:
                await self.async_login()
            except (
                PulseClientConnectionError,
                PulseServerConnectionError,
            ) as ex:
                LOG.log(
                    log_level,
                    "loop login in task %s received exception %s, retrying",
                    task_name,
                    ex.args[0],
                )
                if (
                    log_level == logging.WARNING
                    and self._sync_check_exception is None
                    or self._sync_check_exception != ex
                ):
                    self._set_update_exception(ex)
                continue
            # success, return
            return

    async def _sync_check_task(self) -> None:
        """Asynchronous function that performs a synchronization check task."""

        async def perform_sync_check_query():
            return await self._pulse_connection.async_query(
                ADT_SYNC_CHECK_URI,
                extra_headers={"Sec-Fetch-Mode": "iframe"},
                extra_params={"ts": str(int(time.time() * 1000))},
            )

        task_name = self._get_sync_task_name()
        LOG.debug("creating %s", task_name)

        response_text: str | None = None
        code: int = 200
        have_updates = False
        url: URL | None = None

        def check_sync_check_response() -> bool:
            """
            Validates the sync check response received from the ADT Pulse site.
            Returns:
                bool: True if the sync check response indicates updates, False otherwise

            Raises:
                PulseAccountLockedError if the account is locked and no retry time is available.
                PulseAuthenticationError if the ADT Pulse site returns an authentication error.
                PulseMFAError if the ADT Pulse site returns an MFA error.
                PulseNotLoggedInError if the ADT Pulse site returns a not logged in error.
            """
            if response_text is None:
                LOG.warning("Internal Error: response_text is None")
                return False
            pattern = r"\d+[-]\d+[-]\d+"
            if not re.match(pattern, response_text):
                LOG.warning(
                    "Unexpected sync check format",
                )
                self._pulse_connection.check_login_errors((code, response_text, url))
                return False
            split_text = response_text.split("-")
            if int(split_text[0]) > 9 or int(split_text[1]) > 9:
                return False
            return True

        async def handle_no_updates_exist() -> None:
            if have_updates:
                try:
                    success = await self.async_update()
                except (
                    PulseClientConnectionError,
                    PulseServerConnectionError,
                    PulseGatewayOfflineError,
                ) as e:
                    LOG.debug("Pulse update failed in task %s due to %s", task_name, e)
                    self._set_update_exception(e)
                    return
                except PulseNotLoggedInError:
                    LOG.info(
                        "Pulse update failed in task %s due to not logged in, relogging in...",
                        task_name,
                    )
                    await self._pulse_connection.quick_logout()
                    await self._login_looped(task_name)
                    return
                if not success:
                    LOG.debug("Pulse data update failed in task %s", task_name)
                    return
                self._set_update_exception(None)
            else:
                additional_msg = ""
                if not self.site.gateway.is_online:
                    # bump backoff and resignal since offline and nothing updated
                    self._set_update_exception(
                        PulseGatewayOfflineError(self.site.gateway.backoff)
                    )
                    additional_msg = ", gateway offline so backoff incremented"
                if self._detailed_debug_logging:
                    LOG.debug(
                        "Sync token %s indicates no remote updates to process %s ",
                        response_text,
                        additional_msg,
                    )

        async def shutdown_task(ex: Exception):
            await self._pulse_connection.quick_logout()
            await self._cancel_task(self._timeout_task)
            self._set_update_exception(ex)

        while True:
            try:
                self._sync_check_sleeping.set()
                if not have_updates and not self.site.gateway.is_online:
                    # gateway going back online will trigger a sync check of 1-0-0
                    await self.site.gateway.backoff.wait_for_backoff()
                else:
                    await asyncio.sleep(
                        self.site.gateway.poll_interval if not have_updates else 0.0
                    )
                self._sync_check_sleeping.clear()
                try:
                    code, response_text, url = await perform_sync_check_query()
                except (
                    PulseClientConnectionError,
                    PulseServerConnectionError,
                ) as e:
                    # temporarily unavailble errors should be reported immediately
                    # since the next query will sleep until the retry-after is over
                    msg = ""
                    if e.backoff.backoff_count > WARN_TRANSIENT_FAILURE_THRESHOLD:
                        self._set_update_exception(e)
                    else:
                        msg = ", ignoring..."
                    LOG.debug("Pulse sync check query failed due to %s%s", e, msg)
                    continue
                except (
                    PulseServiceTemporarilyUnavailableError,
                    PulseNotLoggedInError,
                ) as e:
                    if isinstance(e, PulseServiceTemporarilyUnavailableError):
                        status = "temporarily unavailable"
                    else:
                        status = "not logged in"
                    LOG.warning("Pulse service %s, ending %s task", status, task_name)
                    await shutdown_task(e)
                    return
                if not handle_response(
                    code, url, logging.WARNING, "Error querying ADT sync"
                ):
                    continue
                if response_text is None:
                    LOG.warning("Sync check received no response from ADT Pulse site")
                    continue
                more_updates = True
                try:
                    if have_updates:
                        more_updates = check_sync_check_response()
                    else:
                        have_updates = check_sync_check_response()
                except PulseNotLoggedInError:
                    LOG.info("Pulse sync check indicates logged out, re-logging in....")
                    await self._pulse_connection.quick_logout()
                    await self._login_looped(task_name)
                except (
                    PulseAuthenticationError,
                    PulseMFARequiredError,
                    PulseAccountLockedError,
                ) as ex:
                    LOG.error(
                        "Task %s exiting due to error: %s",
                        task_name,
                        ex.args[0],
                    )
                    await shutdown_task(ex)
                    return
                if have_updates and more_updates:
                    LOG.debug("Updates exist: %s, requerying", response_text)
                    continue
                await handle_no_updates_exist()
                have_updates = False
                continue
            except asyncio.CancelledError:
                LOG.debug("%s cancelled", task_name)
                return

    async def async_login(self) -> None:
        """Login asynchronously to ADT.

        Returns: None

        Raises:
            PulseClientConnectionError: if client connection fails
            PulseServerConnectionError: if server connection fails
            PulseServiceTemporarilyUnavailableError: if server returns a Retry-After header
            PulseAuthenticationError: if authentication fails
            PulseAccountLockedError: if account is locked
            PulseMFARequiredError: if MFA is required
            PulseNotLoggedInError: if login fails
        """
        if self._pulse_connection.login_in_progress:
            LOG.debug("Login already in progress, returning")
            return
        LOG.debug(
            "Authenticating to ADT Pulse cloud service as %s",
            self._authentication_properties.username,
        )
        await self._pulse_connection.async_fetch_version()
        soup = await self._pulse_connection.async_do_login_query()
        if soup is None:
            await self._pulse_connection.quick_logout()
            ex = PulseNotLoggedInError()
            self.sync_check_exception = ex
            raise ex
        self.sync_check_exception = None
        # if tasks are started, we've already logged in before
        # clean up completed tasks first
        await self._clean_done_tasks()
        if self._timeout_task is not None:
            return
        if not self._site:
            await self._update_sites(soup)
        if self._site is None:
            LOG.error("Could not retrieve any sites, login failed")
            await self._pulse_connection.quick_logout()
            ex = PulseNotLoggedInError()
            self.sync_check_exception = ex
            raise ex
        self.sync_check_exception = None
        self._timeout_task = asyncio.create_task(
            self._keepalive_task(), name=KEEPALIVE_TASK_NAME
        )
        await asyncio.sleep(0)

    async def async_logout(self) -> None:
        """Logout of ADT Pulse async."""
        if self._pulse_connection.login_in_progress:
            LOG.debug("Login in progress, returning")
            return
        self._set_update_exception(PulseNotLoggedInError())
        LOG.info(
            "Logging %s out of ADT Pulse", self._authentication_properties.username
        )
        if asyncio.current_task() not in (self._sync_task, self._timeout_task):
            await self._cancel_task(self._timeout_task)
            await self._cancel_task(self._sync_task)
        await self._pulse_connection.async_do_logout_query(self.site.id)

    async def async_update(self) -> bool:
        """Update ADT Pulse data.

        Returns:
            bool: True if update succeeded.

        Raises:
            PulseGatewayOfflineError: if the gateway is offline
        """
        LOG.debug("Checking ADT Pulse cloud service for updates")

        # FIXME will have to query other URIs for camera/zwave/etc
        soup = await self._pulse_connection.query_orb(
            logging.INFO, "Error returned from ADT Pulse service check"
        )
        if soup is not None:
            await self._update_sites(soup)
            return True

        return False

    async def wait_for_update(self) -> None:
        """Wait for update.

        Blocks current async task until Pulse system
        signals an update

        Raises:
            Every exception from exceptions.py are possible
        """
        # FIXME?: This code probably won't work with multiple waiters.
        await self._clean_done_tasks()
        if self.sync_check_exception:
            raise self.sync_check_exception
        with self._pa_attribute_lock:
            if self._timeout_task is None:
                raise PulseNotLoggedInError()
            if self._sync_task is None:
                coro = self._sync_check_task()
                self._sync_task = asyncio.create_task(
                    coro, name=f"{SYNC_CHECK_TASK_NAME}: Async session"
                )
                await asyncio.sleep(0)

        await self._pulse_properties.updates_exist.wait()
        self._pulse_properties.updates_exist.clear()
        curr_exception = self.sync_check_exception
        self.sync_check_exception = None
        if curr_exception:
            raise curr_exception

    @property
    def sites(self) -> list[ADTPulseSite]:
        """Return all sites for this ADT Pulse account."""
        warn(
            "multiple sites being removed, use pyADTPulse.site instead",
            PendingDeprecationWarning,
            stacklevel=2,
        )
        with self._pa_attribute_lock:
            if self._site is None:
                raise RuntimeError(
                    "No sites have been retrieved, have you logged in yet?"
                )
            return [self._site]

    @property
    def site(self) -> ADTPulseSite:
        """Return the site associated with the Pulse login."""
        with self._pa_attribute_lock:
            if self._site is None:
                raise RuntimeError(
                    "No sites have been retrieved, have you logged in yet?"
                )
            return self._site

    @property
    def is_connected(self) -> bool:
        """Convenience method to return whether ADT Pulse is connected."""
        return self._pulse_connection.is_connected

    @property
    def detailed_debug_logging(self) -> bool:
        """Return detailed debug logging."""
        return self._pulse_connection.detailed_debug_logging

    @detailed_debug_logging.setter
    @typechecked
    def detailed_debug_logging(self, value: bool) -> None:
        """Set detailed debug logging."""
        self._pulse_connection.detailed_debug_logging = value

    @property
    def keepalive_interval(self) -> int:
        """Get the keepalive interval in minutes.

        Returns:
            int: the keepalive interval
        """
        return self._pulse_properties.keepalive_interval

    @keepalive_interval.setter
    @typechecked
    def keepalive_interval(self, interval: int | None) -> None:
        """Set the keepalive interval in minutes.

        Args:
            interval (int|None): The number of minutes between keepalive calls
                                 If set to None, resets to ADT_DEFAULT_KEEPALIVE_INTERVAL

        Raises:
            ValueError: if a keepalive interval of greater than ADT_MAX_KEEPALIVE_INTERVAL
                        minutes is specified
        """
        self._pulse_properties.keepalive_interval = interval

    @property
    def relogin_interval(self) -> int:
        """Get the relogin interval in minutes.

        Returns:
            int: the relogin interval
        """
        return self._pulse_properties.relogin_interval

    @relogin_interval.setter
    @typechecked
    def relogin_interval(self, interval: int | None) -> None:
        """Set the relogin interval in minutes.

        If set to None, resets to ADT_DEFAULT_RELOGIN_INTERVAL
        """
        self._pulse_properties.relogin_interval = interval

    @property
    def sync_check_exception(self) -> Exception | None:
        """Return sync check exception.

        This should not be used by external code.

        Returns:
            Exception: sync check exception
        """
        with self._pa_attribute_lock:
            return self._sync_check_exception

    @sync_check_exception.setter
    @typechecked
    def sync_check_exception(self, value: Exception | None) -> None:
        """Set sync check exception.

        This should not be used by external code.

        Args:
            value (Exception): sync check exception
        """
        with self._pa_attribute_lock:
            self._sync_check_exception = value



"""Pulse Query Manager."""

from logging import getLogger
from asyncio import wait_for
from datetime import datetime
from http import HTTPStatus
from time import time

from aiohttp import (
    ClientConnectionError,
    ClientConnectorError,
    ClientError,
    ClientResponse,
    ClientResponseError,
    ServerConnectionError,
    ServerDisconnectedError,
    ServerTimeoutError,
)
from bs4 import BeautifulSoup
from typeguard import typechecked
from yarl import URL

from .const import (
    ADT_DEFAULT_LOGIN_TIMEOUT,
    ADT_HTTP_BACKGROUND_URIS,
    ADT_ORB_URI,
    ADT_OTHER_HTTP_ACCEPT_HEADERS,
)
from .exceptions import (
    PulseClientConnectionError,
    PulseNotLoggedInError,
    PulseServerConnectionError,
    PulseServiceTemporarilyUnavailableError,
)
from .pulse_backoff import PulseBackoff
from .pulse_connection_properties import PulseConnectionProperties
from .pulse_connection_status import PulseConnectionStatus
from .util import make_soup, set_debug_lock

LOG = getLogger(__name__)

RECOVERABLE_ERRORS = {
    HTTPStatus.INTERNAL_SERVER_ERROR,
    HTTPStatus.BAD_GATEWAY,
    HTTPStatus.GATEWAY_TIMEOUT,
}

MAX_REQUERY_RETRIES = 3


class PulseQueryManager:
    """Pulse Query Manager."""

    __slots__ = (
        "_pqm_attribute_lock",
        "_connection_properties",
        "_connection_status",
        "_debug_locks",
    )

    @staticmethod
    @typechecked
    def _get_http_status_description(status_code: int) -> str:
        """Get HTTP status description."""
        status = HTTPStatus(status_code)
        return status.description

    @typechecked
    def __init__(
        self,
        connection_status: PulseConnectionStatus,
        connection_properties: PulseConnectionProperties,
        debug_locks: bool = False,
    ) -> None:
        """Initialize Pulse Query Manager."""
        self._pqm_attribute_lock = set_debug_lock(
            debug_locks, "pyadtpulse.pqm_attribute_lock"
        )
        self._connection_status = connection_status
        self._connection_properties = connection_properties
        self._debug_locks = debug_locks

    @staticmethod
    @typechecked
    async def _handle_query_response(
        response: ClientResponse | None,
    ) -> tuple[int, str | None, URL | None, str | None]:
        if response is None:
            return 0, None, None, None
        response_text = await response.text()

        return (
            response.status,
            response_text,
            response.url,
            response.headers.get("Retry-After"),
        )

    @typechecked
    def _handle_http_errors(
        self, return_value: tuple[int, str | None, URL | None, str | None]
    ) -> None:
        """Handle HTTP errors.

        Parameters:
            return_value (tuple[int, str | None, URL | None, str | None]):
                The return value from _handle_query_response.

        Raises:
            PulseServerConnectionError: If the server returns an error code.
            PulseServiceTemporarilyUnavailableError: If the server returns a
                Retry-After header."""

        def get_retry_after(retry_after: str) -> int | None:
            """
            Parse the value of the "Retry-After" header.

            Parameters:
                retry_after (str): The value of the "Retry-After" header

            Returns:
                int | None: The timestamp in seconds to wait before retrying, or None if the header is invalid.
            """
            if retry_after.isnumeric():
                retval = int(retry_after) + int(time())
            else:
                try:
                    retval = int(
                        datetime.strptime(
                            retry_after, "%a, %d %b %Y %H:%M:%S %Z"
                        ).timestamp()
                    )
                except ValueError:
                    return None
            return retval

        if return_value[0] in (
            HTTPStatus.TOO_MANY_REQUESTS,
            HTTPStatus.SERVICE_UNAVAILABLE,
        ):
            retry = None
            if return_value[3]:
                retry = get_retry_after(return_value[3])
            raise PulseServiceTemporarilyUnavailableError(
                self._connection_status.get_backoff(),
                retry,
            )
        raise PulseServerConnectionError(
            f"HTTP error {return_value[0]}: {return_value[1]} connecting to {return_value[2]}",
            self._connection_status.get_backoff(),
        )

    @typechecked
    def _handle_network_errors(self, e: Exception) -> None:
        if type(e) in (
            ServerConnectionError,
            ServerTimeoutError,
            ServerDisconnectedError,
        ):
            raise PulseServerConnectionError(
                str(e), self._connection_status.get_backoff()
            )
        if (
            isinstance(e, ClientConnectionError)
            and "Connection refused" in str(e)
            or ("timed out") in str(e)
        ):
            raise PulseServerConnectionError(
                str(e), self._connection_status.get_backoff()
            )
        if isinstance(e, ClientConnectorError) and e.os_error not in (
            TimeoutError,
            BrokenPipeError,
        ):
            raise PulseServerConnectionError(
                str(e), self._connection_status.get_backoff()
            )
        raise PulseClientConnectionError(str(e), self._connection_status.get_backoff())

    @typechecked
    async def async_query(
        self,
        uri: str,
        method: str = "GET",
        extra_params: dict[str, str] | None = None,
        extra_headers: dict[str, str] | None = None,
        timeout: int = 1,
        requires_authentication: bool = True,
    ) -> tuple[int, str | None, URL | None]:
        """
        Query ADT Pulse async.

        Args:
            uri (str): URI to query
            method (str, optional): method to use. Defaults to "GET".
            extra_params (Optional[Dict], optional): query/body parameters.
                                                    Defaults to None.
            extra_headers (Optional[Dict], optional): extra HTTP headers.
                                                    Defaults to None.
            timeout (int, optional): timeout in seconds. Defaults to 1.
            requires_authentication (bool, optional): True if authentication is
                                                    required to perform query.
                                                    Defaults to True.
                                                    If true and authenticated flag not
                                                    set, will wait for flag to be set.

        Returns:
            tuple with integer return code, optional response text, and optional URL of
            response

        Raises:
            PulseClientConnectionError: If the client cannot connect
            PulseServerConnectionError: If there is a server error
            PulseServiceTemporarilyUnavailableError: If the server returns a Retry-After header
            PulseNotLoggedInError: if not logged in and task is waiting for longer than
                ADT_DEFAULT_LOGIN_TIMEOUT seconds
        """

        async def setup_query():
            if method not in ("GET", "POST"):
                raise ValueError("method must be GET or POST")
            await self._connection_status.get_backoff().wait_for_backoff()
            if not self._connection_properties.api_version:
                await self.async_fetch_version()
                if not self._connection_properties.api_version:
                    raise ValueError("Could not determine API version for connection")

        retry_after = self._connection_status.retry_after
        now = time()
        if retry_after > now:
            raise PulseServiceTemporarilyUnavailableError(
                self._connection_status.get_backoff(), retry_after
            )
        await setup_query()
        url = self._connection_properties.make_url(uri)
        headers = extra_headers if extra_headers is not None else {}
        if uri in ADT_HTTP_BACKGROUND_URIS:
            headers.setdefault("Accept", ADT_OTHER_HTTP_ACCEPT_HEADERS["Accept"])
        if self._connection_properties.detailed_debug_logging:
            LOG.debug(
                "Attempting %s %s params=%s timeout=%d",
                method,
                url,
                extra_params,
                timeout,
            )
        retry = 0
        return_value: tuple[int, str | None, URL | None, str | None] = (
            HTTPStatus.OK.value,
            None,
            None,
            None,
        )
        query_backoff = PulseBackoff(
            f"Query:{method} {uri}",
            self._connection_status.get_backoff().initial_backoff_interval,
            threshold=0,
            debug_locks=self._debug_locks,
            detailed_debug_logging=self._connection_properties.detailed_debug_logging,
        )
        max_retries = (
            MAX_REQUERY_RETRIES
            if not self._connection_status.get_backoff().will_backoff()
            else 1
        )
        while retry < max_retries:
            try:
                await query_backoff.wait_for_backoff()
                retry += 1
                if (
                    requires_authentication
                    and not self._connection_status.authenticated_flag.is_set()
                ):
                    if self._connection_properties.detailed_debug_logging:
                        LOG.debug(
                            "%s for %s waiting for authenticated flag to be set",
                            method,
                            uri,
                        )
                    # wait for authenticated flag to be set
                    # use a timeout to prevent waiting forever
                    try:
                        await wait_for(
                            self._connection_status.authenticated_flag.wait(),
                            ADT_DEFAULT_LOGIN_TIMEOUT,
                        )
                    except TimeoutError as ex:
                        LOG.warning(
                            "%s for %s timed out waiting for authenticated flag to be set",
                            method,
                            uri,
                        )
                        raise PulseNotLoggedInError() from ex
                async with self._connection_properties.session.request(
                    method,
                    url,
                    headers=extra_headers,
                    params=extra_params if method == "GET" else None,
                    data=extra_params if method == "POST" else None,
                    timeout=timeout,
                ) as response:
                    return_value = await self._handle_query_response(response)
                    if return_value[0] in RECOVERABLE_ERRORS:
                        LOG.debug(
                            "query returned recoverable error code %s: %s,"
                            "retrying (count = %d)",
                            return_value[0],
                            self._get_http_status_description(return_value[0]),
                            retry,
                        )
                        if max_retries > 1 and retry == max_retries:
                            LOG.debug(
                                "Exceeded max retries of %d, giving up", max_retries
                            )
                        else:
                            query_backoff.increment_backoff()
                            response.raise_for_status()
                        continue
                    response.raise_for_status()
                    break

            except ClientResponseError:
                self._handle_http_errors(return_value)
            except (
                ClientConnectorError,
                ServerTimeoutError,
                ClientError,
                ServerConnectionError,
                ServerDisconnectedError,
            ) as ex:
                LOG.debug(
                    "Error %s occurred making %s request to %s",
                    ex.args,
                    method,
                    url,
                    exc_info=True,
                )
                if retry == max_retries:
                    self._handle_network_errors(ex)
                query_backoff.increment_backoff()
                continue
            except TimeoutError as ex:
                if retry == max_retries:
                    LOG.debug("Exceeded max retries of %d, giving up", max_retries)
                    raise PulseServerConnectionError(
                        "Timeout error",
                        self._connection_status.get_backoff(),
                    ) from ex
                query_backoff.increment_backoff()
                continue
        # success
        self._connection_status.get_backoff().reset_backoff()
        return (return_value[0], return_value[1], return_value[2])

    async def query_orb(self, level: int, error_message: str) -> BeautifulSoup | None:
        """Query ADT Pulse ORB.

        Args:
            level (int): error level to log on failure
            error_message (str): error message to use on failure

        Returns:
            Optional[BeautifulSoup]: A Beautiful Soup object, or None if failure

        Raises:
            PulseClientConnectionError: If the client cannot connect
            PulseServerConnectionError: If there is a server error
            PulseServiceTemporarilyUnavailableError: If the server returns a Retry-After header
        """
        code, response, url = await self.async_query(
            ADT_ORB_URI,
            extra_headers={"Sec-Fetch-Mode": "cors", "Sec-Fetch-Dest": "empty"},
        )

        return make_soup(code, response, url, level, error_message)

    async def async_fetch_version(self) -> None:
        """Fetch ADT Pulse version.

        Exceptions are passed through to the caller since if this fails, there is
        probably some underlying connection issue.
        """
        response_values: tuple[int, str | None, URL | None, str | None] = (
            HTTPStatus.OK.value,
            None,
            None,
            None,
        )
        if self._connection_properties.api_version:
            return

        signin_url = self._connection_properties.service_host
        try:
            async with self._connection_properties.session.get(
                signin_url, timeout=10
            ) as response:
                response_values = await self._handle_query_response(response)
                response.raise_for_status()

        except ClientResponseError as ex:
            LOG.error(
                "Error %s occurred determining Pulse API version",
                ex.args,
                exc_info=True,
            )
            self._handle_http_errors(response_values)
            return
        except (
            ClientConnectorError,
            ServerTimeoutError,
            ClientError,
            ServerConnectionError,
        ) as ex:
            LOG.error(
                "Error %s occurred determining Pulse API version",
                ex.args,
                exc_info=True,
            )
            self._handle_network_errors(ex)
        except TimeoutError as ex:
            LOG.error(
                "Timeout occurred determining Pulse API version %s",
                ex.args,
                exc_info=True,
            )
            raise PulseServerConnectionError(
                "Timeout occurred determining Pulse API version",
                self._connection_status.get_backoff(),
            ) from ex
        version = self._connection_properties.get_api_version(str(response_values[2]))
        if version is not None:
            self._connection_properties.api_version = version
            LOG.debug(
                "Discovered ADT Pulse version %s at %s",
                self._connection_properties.api_version,
                self._connection_properties.service_host,
            )
            self._connection_status.get_backoff().reset_backoff()



"""Pulse backoff object."""

import asyncio
import datetime
from logging import getLogger
from time import time

from typeguard import typechecked

from .const import ADT_MAX_BACKOFF
from .util import set_debug_lock

LOG = getLogger(__name__)


class PulseBackoff:
    """Pulse backoff object."""

    __slots__ = (
        "_b_lock",
        "_initial_backoff_interval",
        "_max_backoff_interval",
        "_backoff_count",
        "_expiration_time",
        "_name",
        "_detailed_debug_logging",
        "_threshold",
    )

    @typechecked
    def __init__(
        self,
        name: str,
        initial_backoff_interval: float,
        max_backoff_interval: float = ADT_MAX_BACKOFF,
        threshold: int = 0,
        debug_locks: bool = False,
        detailed_debug_logging=False,
    ) -> None:
        """Initialize backoff.

        Args:
            name (str): Name of the backoff.
            initial_backoff_interval (float): Initial backoff interval in seconds.
            max_backoff_interval (float, optional): Maximum backoff interval in seconds.
                Defaults to ADT_MAX_BACKOFF.
            threshold (int, optional): Threshold for backoff. Defaults to 0.
            debug_locks (bool, optional): Enable debug locks. Defaults to False.
            detailed_debug_logging (bool, optional): Enable detailed debug logging.
                Defaults to False.
        """
        self._check_intervals(initial_backoff_interval, max_backoff_interval)
        self._b_lock = set_debug_lock(debug_locks, "pyadtpulse._b_lock")
        self._initial_backoff_interval = initial_backoff_interval
        self._max_backoff_interval = max_backoff_interval
        self._backoff_count = 0
        self._expiration_time = 0.0
        self._name = name
        self._detailed_debug_logging = detailed_debug_logging
        self._threshold = threshold

    def _calculate_backoff_interval(self) -> float:
        """Calculate backoff time."""
        if self._backoff_count == 0:
            return 0.0
        if self._backoff_count <= (self._threshold + 1):
            return self._initial_backoff_interval
        return min(
            self._initial_backoff_interval
            * 2 ** (self._backoff_count - self._threshold - 1),
            self._max_backoff_interval,
        )

    @staticmethod
    def _check_intervals(
        initial_backoff_interval: float, max_backoff_interval: float
    ) -> None:
        """Check max_backoff_interval is >= initial_backoff_interval
        and that both invervals are positive."""
        if initial_backoff_interval <= 0:
            raise ValueError("initial_backoff_interval must be greater than 0")
        if max_backoff_interval < initial_backoff_interval:
            raise ValueError("max_backoff_interval must be >= initial_backoff_interval")

    def get_current_backoff_interval(self) -> float:
        """Return current backoff time."""
        with self._b_lock:
            return self._calculate_backoff_interval()

    def increment_backoff(self) -> None:
        """Increment backoff."""
        with self._b_lock:
            self._backoff_count += 1
            if self._detailed_debug_logging:
                LOG.debug(
                    "Pulse backoff %s: incremented to %s",
                    self._name,
                    self._backoff_count,
                )

    def reset_backoff(self) -> None:
        """Reset backoff."""
        with self._b_lock:
            if self._expiration_time < time():
                if self._detailed_debug_logging and self._backoff_count != 0:
                    LOG.debug("Pulse backoff %s reset", self._name)
                self._backoff_count = 0
                self._expiration_time = 0.0

    @typechecked
    def set_absolute_backoff_time(self, backoff_time: float) -> None:
        """Set absolute backoff time."""
        curr_time = time()
        if backoff_time < curr_time:
            raise ValueError("Absolute backoff time must be greater than current time")
        with self._b_lock:
            if self._detailed_debug_logging:
                LOG.debug(
                    "Pulse backoff %s: set to %s",
                    self._name,
                    datetime.datetime.fromtimestamp(backoff_time).strftime(
                        "%m/%d/%Y %H:%M:%S"
                    ),
                )
            self._expiration_time = backoff_time
            self._backoff_count = 0

    async def wait_for_backoff(self) -> None:
        """Wait for backoff."""
        with self._b_lock:
            curr_time = time()
            if self._expiration_time < curr_time:
                if self.backoff_count == 0:
                    return
                diff = self._calculate_backoff_interval()
            else:
                diff = self._expiration_time - curr_time
            if diff > 0:
                if self._detailed_debug_logging:
                    LOG.debug("Backoff %s: waiting for %s", self._name, diff)
                await asyncio.sleep(diff)

    def will_backoff(self) -> bool:
        """Return if backoff is needed."""
        with self._b_lock:
            return (
                self._backoff_count > self._threshold or self._expiration_time >= time()
            )

    @property
    def backoff_count(self) -> int:
        """Return backoff count."""
        with self._b_lock:
            return self._backoff_count

    @property
    def expiration_time(self) -> float:
        """Return backoff expiration time."""
        with self._b_lock:
            return self._expiration_time

    @property
    def initial_backoff_interval(self) -> float:
        """Return initial backoff interval."""
        with self._b_lock:
            return self._initial_backoff_interval

    @initial_backoff_interval.setter
    @typechecked
    def initial_backoff_interval(self, new_interval: float) -> None:
        """Set initial backoff interval."""
        with self._b_lock:
            self._check_intervals(new_interval, self._max_backoff_interval)
            self._initial_backoff_interval = new_interval

    @property
    def name(self) -> str:
        """Return name."""
        return self._name

    @property
    def detailed_debug_logging(self) -> bool:
        """Return detailed debug logging."""
        with self._b_lock:
            return self._detailed_debug_logging

    @detailed_debug_logging.setter
    @typechecked
    def detailed_debug_logging(self, new_value: bool) -> None:
        """Set detailed debug logging."""
        with self._b_lock:
            self._detailed_debug_logging = new_value



"""ADT Pulse connection. End users should probably not call this directly.

This is the main interface to the http functions to access ADT Pulse.
"""

import logging
import re
from asyncio import AbstractEventLoop
from time import time

from bs4 import BeautifulSoup
from typeguard import typechecked
from yarl import URL

from .const import (
    ADT_DEFAULT_LOGIN_TIMEOUT,
    ADT_LOGIN_URI,
    ADT_LOGOUT_URI,
    ADT_MFA_FAIL_URI,
    ADT_SUMMARY_URI,
)
from .exceptions import (
    PulseAccountLockedError,
    PulseAuthenticationError,
    PulseClientConnectionError,
    PulseMFARequiredError,
    PulseNotLoggedInError,
    PulseServerConnectionError,
    PulseServiceTemporarilyUnavailableError,
)
from .pulse_authentication_properties import PulseAuthenticationProperties
from .pulse_backoff import PulseBackoff
from .pulse_connection_properties import PulseConnectionProperties
from .pulse_connection_status import PulseConnectionStatus
from .pulse_query_manager import PulseQueryManager
from .util import make_soup, set_debug_lock

LOG = logging.getLogger(__name__)


SESSION_COOKIES = {"X-mobile-browser": "false", "ICLocal": "en_US"}


class PulseConnection(PulseQueryManager):
    """ADT Pulse connection related attributes."""

    __slots__ = (
        "_pc_attribute_lock",
        "_authentication_properties",
        "_login_backoff",
        "_login_in_progress",
    )

    @typechecked
    def __init__(
        self,
        pulse_connection_status: PulseConnectionStatus,
        pulse_connection_properties: PulseConnectionProperties,
        pulse_authentication: PulseAuthenticationProperties,
        debug_locks: bool = False,
    ):
        """Initialize ADT Pulse connection."""

        # need to initialize this after the session since we set cookies
        # based on it
        super().__init__(
            pulse_connection_status,
            pulse_connection_properties,
            debug_locks,
        )
        self._pc_attribute_lock = set_debug_lock(
            debug_locks, "pyadtpulse.pc_attribute_lock"
        )
        self._connection_properties = pulse_connection_properties
        self._connection_status = pulse_connection_status
        self._authentication_properties = pulse_authentication
        self._login_backoff = PulseBackoff(
            "Login",
            pulse_connection_status._backoff.initial_backoff_interval,
            detailed_debug_logging=self._connection_properties.detailed_debug_logging,
        )
        self._login_in_progress = False
        self._debug_locks = debug_locks

    @typechecked
    def check_login_errors(
        self, response: tuple[int, str | None, URL | None]
    ) -> BeautifulSoup:
        """Check response for login errors.

        Will handle setting backoffs and raising exceptions.

        Args:
            response (tuple[int, str | None, URL | None]): The response

        Returns:
            BeautifulSoup: The parsed response

        Raises:
            PulseAuthenticationError: if login fails due to incorrect username/password
            PulseServerConnectionError: if login fails due to server error
            PulseAccountLockedError: if login fails due to account locked
            PulseMFARequiredError: if login fails due to MFA required
            PulseNotLoggedInError: if login fails due to not logged in
        """

        def extract_seconds_from_string(s: str) -> int:
            seconds = 0
            match = re.search(r"\d+", s)
            if match:
                seconds = int(match.group())
                if "minutes" in s:
                    seconds *= 60
            return seconds

        def determine_error_type():
            """Determine what type of error we have from the url and the parsed page.

            Will raise the appropriate exception.
            """
            self._login_in_progress = False
            url = self._connection_properties.make_url(ADT_LOGIN_URI)
            if url == response_url_string:
                error = soup.find("div", {"id": "warnMsgContents"})
                if error:
                    error_text = error.get_text()
                    LOG.error("Error logging into pulse: %s", error_text)
                    if "Try again in" in error_text:
                        if (retry_after := extract_seconds_from_string(error_text)) > 0:
                            raise PulseAccountLockedError(
                                self._login_backoff,
                                retry_after + time(),
                            )
                    elif "You have not yet signed in" in error_text:
                        raise PulseNotLoggedInError()
                    elif "Sign In Unsuccessful" in error_text:
                        raise PulseAuthenticationError()
                else:
                    raise PulseNotLoggedInError()
            else:
                url = self._connection_properties.make_url(ADT_MFA_FAIL_URI)
                if url == response_url_string:
                    raise PulseMFARequiredError()

        soup = make_soup(
            response[0],
            response[1],
            response[2],
            logging.ERROR,
            "Could not log into ADT Pulse site",
        )
        # this probably should have been handled by async_query()
        if soup is None:
            raise PulseServerConnectionError(
                f"Could not log into ADT Pulse site: code {response[0]}: URL: {response[2]}, response: {response[1]}",
                self._login_backoff,
            )
        url = self._connection_properties.make_url(ADT_SUMMARY_URI)
        response_url_string = str(response[2])
        if url != response_url_string:
            determine_error_type()
            raise PulseAuthenticationError()
        return soup

    @typechecked
    async def async_do_login_query(
        self, timeout: int = ADT_DEFAULT_LOGIN_TIMEOUT
    ) -> BeautifulSoup | None:
        """
        Performs a login query to the Pulse site.

        Will backoff on login failures.

        Will set login in progress flag.

        Args:
            timeout (int, optional): The timeout value for the query in seconds.
            Defaults to ADT_DEFAULT_LOGIN_TIMEOUT.

        Returns:
            soup: Optional[BeautifulSoup]: A BeautifulSoup object containing
            summary.jsp, or None if failure
        Raises:
            ValueError: if login parameters are not correct
            PulseAuthenticationError: if login fails due to incorrect username/password
            PulseServerConnectionError: if login fails due to server error
            PulseServiceTemporarilyUnavailableError: if login fails due to too many requests or
                server is temporarily unavailable
            PulseAccountLockedError: if login fails due to account locked
            PulseMFARequiredError: if login fails due to MFA required
            PulseNotLoggedInError: if login fails due to not logged in (which is probably an internal error)
        """

        if self.login_in_progress:
            return None
        await self.quick_logout()
        # just raise exceptions if we're not going to be able to log in
        lockout_time = self._login_backoff.expiration_time
        if lockout_time > time():
            raise PulseAccountLockedError(self._login_backoff, lockout_time)
        cs_backoff = self._connection_status.get_backoff()
        lockout_time = cs_backoff.expiration_time
        if lockout_time > time():
            raise PulseServiceTemporarilyUnavailableError(cs_backoff, lockout_time)
        self.login_in_progress = True
        data = {
            "usernameForm": self._authentication_properties.username,
            "passwordForm": self._authentication_properties.password,
            "networkid": self._authentication_properties.site_id,
            "fingerprint": self._authentication_properties.fingerprint,
        }
        await self._login_backoff.wait_for_backoff()
        try:
            response = await self.async_query(
                ADT_LOGIN_URI,
                "POST",
                extra_params=data,
                timeout=timeout,
                requires_authentication=False,
            )
        except (
            PulseClientConnectionError,
            PulseServerConnectionError,
            PulseServiceTemporarilyUnavailableError,
        ) as e:
            LOG.error("Could not log into Pulse site: %s", e)
            self.login_in_progress = False
            raise
        soup = self.check_login_errors(response)
        self._connection_status.authenticated_flag.set()
        self._authentication_properties.last_login_time = int(time())
        self._login_backoff.reset_backoff()
        self.login_in_progress = False
        return soup

    @typechecked
    async def async_do_logout_query(self, site_id: str | None = None) -> None:
        """Performs a logout query to the ADT Pulse site."""
        params = {}
        si = ""
        self._connection_status.authenticated_flag.clear()
        if site_id is not None and site_id != "":
            self._authentication_properties.site_id = site_id
            si = site_id
        params.update({"networkid": si})

        params.update({"partner": "adt"})
        try:
            await self.async_query(
                ADT_LOGOUT_URI,
                extra_params=params,
                timeout=10,
                requires_authentication=False,
            )
        # FIXME: do we care if this raises exceptions?
        except (
            PulseClientConnectionError,
            PulseServiceTemporarilyUnavailableError,
            PulseServerConnectionError,
        ) as e:
            LOG.debug("Could not logout from Pulse site: %s", e)

    @property
    def is_connected(self) -> bool:
        """Check if ADT Pulse is connected."""
        return (
            self._connection_status.authenticated_flag.is_set()
            and not self._login_in_progress
        )

    @property
    def login_backoff(self) -> PulseBackoff:
        """Return backoff object."""
        with self._pc_attribute_lock:
            return self._login_backoff

    def check_sync(self, message: str) -> AbstractEventLoop:
        """Convenience method to check if running from sync context."""
        return self._connection_properties.check_sync(message)

    @property
    def debug_locks(self):
        """Return debug locks."""
        return self._debug_locks

    @property
    def login_in_progress(self) -> bool:
        """Return login in progress."""
        with self._pc_attribute_lock:
            return self._login_in_progress

    @login_in_progress.setter
    @typechecked
    def login_in_progress(self, value: bool) -> None:
        """Set login in progress."""
        with self._pc_attribute_lock:
            self._login_in_progress = value

    async def quick_logout(self) -> None:
        """Quickly logout.

        This just resets the authenticated flag and clears the ClientSession.
        """
        LOG.debug("Resetting session")
        self._connection_status.authenticated_flag.clear()
        await self._connection_properties.clear_session()

    @property
    def detailed_debug_logging(self) -> bool:
        """Return detailed debug logging."""
        return (
            self._login_backoff.detailed_debug_logging
            and self._connection_properties.detailed_debug_logging
            and self._connection_status.detailed_debug_logging
        )

    @detailed_debug_logging.setter
    @typechecked
    def detailed_debug_logging(self, value: bool):
        with self._pc_attribute_lock:
            self._login_backoff.detailed_debug_logging = value
            self._connection_properties.detailed_debug_logging = value
            self._connection_status.detailed_debug_logging = value

    def get_login_backoff(self) -> PulseBackoff:
        """Return login backoff."""
        return self._login_backoff



"""Pulse Authentication Properties."""

from re import match

from typeguard import typechecked

from .util import set_debug_lock


class PulseAuthenticationProperties:
    """Pulse Authentication Properties."""

    __slots__ = (
        "_username",
        "_password",
        "_fingerprint",
        "_paa_attribute_lock",
        "_last_login_time",
        "_site_id",
    )

    @staticmethod
    def check_username(username: str) -> None:
        """Check if username is valid.

        Raises ValueError if a login parameter is not valid."""
        if not username:
            raise ValueError("Username is mandatory")
        pattern = r"\b[A-Za-z0-9._%+-]+@[A-Za-z0-9.-]+\.[A-Z|a-z]{2,}\b"
        if not match(pattern, username):
            raise ValueError("Username must be an email address")

    @staticmethod
    @typechecked
    def check_password(password: str) -> None:
        """Check if password is valid.

        Raises ValueError if password is not valid.
        """
        if not password:
            raise ValueError("Password is mandatory")

    @staticmethod
    @typechecked
    def check_fingerprint(fingerprint: str) -> None:
        """Check if fingerprint is valid.

        Raises ValueError if password is not valid.
        """
        if not fingerprint:
            raise ValueError("Fingerprint is required")

    @typechecked
    def __init__(
        self,
        username: str,
        password: str,
        fingerprint: str,
        debug_locks: bool = False,
    ) -> None:
        """Initialize Pulse Authentication Properties."""
        self.check_username(username)
        self.check_password(password)
        self.check_fingerprint(fingerprint)
        self._username = username
        self._password = password
        self._fingerprint = fingerprint
        self._paa_attribute_lock = set_debug_lock(
            debug_locks, "pyadtpulse.paa_attribute_lock"
        )
        self._last_login_time = 0
        self._site_id = ""

    @property
    def last_login_time(self) -> int:
        """Get the last login time."""
        with self._paa_attribute_lock:
            return self._last_login_time

    @last_login_time.setter
    @typechecked
    def last_login_time(self, login_time: int) -> None:
        with self._paa_attribute_lock:
            self._last_login_time = login_time

    @property
    def username(self) -> str:
        """Get the username."""
        with self._paa_attribute_lock:
            return self._username

    @username.setter
    @typechecked
    def username(self, username: str) -> None:
        self.check_username(username)
        with self._paa_attribute_lock:
            self._username = username

    @property
    def password(self) -> str:
        """Get the password."""
        with self._paa_attribute_lock:
            return self._password

    @password.setter
    @typechecked
    def password(self, password: str) -> None:
        self.check_password(password)
        with self._paa_attribute_lock:
            self._password = password

    @property
    def fingerprint(self) -> str:
        """Get the fingerprint."""
        with self._paa_attribute_lock:
            return self._fingerprint

    @fingerprint.setter
    @typechecked
    def fingerprint(self, fingerprint: str) -> None:
        self.check_fingerprint(fingerprint)
        with self._paa_attribute_lock:
            self._fingerprint = fingerprint

    @property
    def site_id(self) -> str:
        """Get the site ID."""
        with self._paa_attribute_lock:
            return self._site_id

    @site_id.setter
    @typechecked
    def site_id(self, site_id: str) -> None:
        with self._paa_attribute_lock:
            self._site_id = site_id



"""Pulse exceptions."""

import datetime
from time import time

from .pulse_backoff import PulseBackoff


def compute_retry_time(retry_time: float | None) -> str:
    """Compute the retry time."""
    if not retry_time:
        return "indefinitely"
    return str(datetime.datetime.fromtimestamp(retry_time))


class PulseExceptionWithBackoff(Exception):
    """Exception with backoff."""

    def __init__(self, message: str, backoff: PulseBackoff):
        """Initialize exception."""
        super().__init__(message)
        self.backoff = backoff
        self.backoff.increment_backoff()

    def __str__(self):
        """Return a string representation of the exception."""
        return f"{self.__class__.__name__}: {self.args[0]}"

    def __repr__(self):
        """Return a string representation of the exception."""
        return f"{self.__class__.__name__}(message='{self.args[0]}', backoff={self.backoff})"


class PulseExceptionWithRetry(PulseExceptionWithBackoff):
    """Exception with backoff

    If retry_time is None, or is in the past, then just the backoff count will be incremented.
    """

    def __init__(self, message: str, backoff: PulseBackoff, retry_time: float | None):
        """Initialize exception."""
        # super.__init__ will increment the backoff count
        super().__init__(message, backoff)
        self.retry_time = retry_time
        if retry_time and retry_time > time():
            # set the absolute backoff time will remove the backoff count
            self.backoff.set_absolute_backoff_time(retry_time)
            return

    def __str__(self):
        """Return a string representation of the exception."""
        return f"{self.__class__.__name__}: {self.args[0]}"

    def __repr__(self):
        """Return a string representation of the exception."""
        return f"{self.__class__.__name__}(message='{self.args[0]}', backoff={self.backoff}, retry_time={self.retry_time})"


class PulseConnectionError(Exception):
    """Base class for connection errors"""


class PulseServerConnectionError(PulseExceptionWithBackoff, PulseConnectionError):
    """Server error."""

    def __init__(self, message: str, backoff: PulseBackoff):
        """Initialize Pulse server error exception."""
        super().__init__(f"Pulse server error: {message}", backoff)


class PulseClientConnectionError(PulseExceptionWithBackoff, PulseConnectionError):
    """Client error."""

    def __init__(self, message: str, backoff: PulseBackoff):
        """Initialize Pulse client error exception."""
        super().__init__(f"Client error connecting to Pulse: {message}", backoff)


class PulseServiceTemporarilyUnavailableError(
    PulseExceptionWithRetry, PulseConnectionError
):
    """Service temporarily unavailable error.

    For HTTP 503 and 429 errors.
    """

    def __init__(self, backoff: PulseBackoff, retry_time: float | None = None):
        """Initialize Pusle service temporarily unavailable error exception."""
        super().__init__(
            f"Pulse service temporarily unavailable until {compute_retry_time(retry_time)}",
            backoff,
            retry_time,
        )


class PulseLoginException(Exception):
    """Login exceptions.

    Base class for catching all login exceptions."""


class PulseAuthenticationError(PulseLoginException):
    """Authentication error."""

    def __init__(self):
        """Initialize Pulse Authentication error exception."""
        super().__init__("Error authenticating to Pulse")


class PulseAccountLockedError(PulseExceptionWithRetry, PulseLoginException):
    """Account locked error."""

    def __init__(self, backoff: PulseBackoff, retry: float):
        """Initialize Pulse Account locked error exception."""
        super().__init__(
            f"Pulse Account is locked until {compute_retry_time(retry)}", backoff, retry
        )


class PulseGatewayOfflineError(PulseExceptionWithBackoff):
    """Gateway offline error."""

    def __init__(self, backoff: PulseBackoff):
        """Initialize Pulse Gateway offline error exception."""
        super().__init__("Gateway is offline", backoff)


class PulseMFARequiredError(PulseLoginException):
    """MFA required error."""

    def __init__(self):
        """Initialize Pulse MFA required error exception."""
        super().__init__("Authentication failed because MFA is required")


class PulseNotLoggedInError(PulseLoginException):
    """Exception to indicate that the application code is not logged in.

    Used for signalling waiters.
    """

    def __init__(self):
        """Initialize Pulse Not logged in error exception."""
        super().__init__("Not logged into Pulse")



"""ADT Pulse Gateway Dataclass."""

import logging
import re
from dataclasses import dataclass
from ipaddress import IPv4Address, IPv6Address, ip_address
from threading import RLock
from typing import Any

from typeguard import typechecked

from .const import ADT_DEFAULT_POLL_INTERVAL, ADT_GATEWAY_MAX_OFFLINE_POLL_INTERVAL
from .pulse_backoff import PulseBackoff
from .util import parse_pulse_datetime

LOG = logging.getLogger(__name__)

STRING_UPDATEABLE_FIELDS = (
    "manufacturer",
    "model",
    "serial_number",
    "firmware_version",
    "hardware_version",
    "primary_connection_type",
    "broadband_connection_status",
    "cellular_connection_status",
    "cellular_connection_signal_strength",
    "broadband_lan_mac",
    "device_lan_mac",
)

DATETIME_UPDATEABLE_FIELDS = ("next_update", "last_update")

IPADDR_UPDATEABLE_FIELDS = (
    "broadband_lan_ip_address",
    "device_lan_ip_address",
    "router_lan_ip_address",
    "router_wan_ip_address",
)


@dataclass(slots=True)
class ADTPulseGateway:
    """ADT Pulse Gateway information."""

    manufacturer: str = "Unknown"
    _status_text: str = "OFFLINE"
    backoff = PulseBackoff(
        "Gateway", ADT_DEFAULT_POLL_INTERVAL, ADT_GATEWAY_MAX_OFFLINE_POLL_INTERVAL
    )
    _attribute_lock = RLock()
    model: str | None = None
    serial_number: str | None = None
    next_update: int = 0
    last_update: int = 0
    firmware_version: str | None = None
    hardware_version: str | None = None
    primary_connection_type: str | None = None
    broadband_connection_status: str | None = None
    cellular_connection_status: str | None = None
    _cellular_connection_signal_strength: float = 0.0
    broadband_lan_ip_address: IPv4Address | IPv6Address | None = None
    _broadband_lan_mac: str | None = None
    device_lan_ip_address: IPv4Address | IPv6Address | None = None
    _device_lan_mac: str | None = None
    router_lan_ip_address: IPv4Address | IPv6Address | None = None
    router_wan_ip_address: IPv4Address | IPv6Address | None = None

    @property
    def is_online(self) -> bool:
        """Returns whether gateway is online.

        Returns:
            bool: True if gateway is online
        """
        with self._attribute_lock:
            return self._status_text == "ONLINE"

    @is_online.setter
    @typechecked
    def is_online(self, status: bool) -> None:
        """Set gateway status.

        Args:
            status (bool): True if gateway is online
        """
        with self._attribute_lock:
            if status == self.is_online:
                return
            old_status = self._status_text
            self._status_text = "ONLINE"
            if not status:
                self._status_text = "OFFLINE"

            LOG.info(
                "ADT Pulse gateway %s",
                self._status_text,
            )
            if old_status == "OFFLINE":
                self.backoff.reset_backoff()
            LOG.debug(
                "Gateway poll interval: %d",
                (
                    self.backoff.initial_backoff_interval
                    if self._status_text == "ONLINE"
                    else self.backoff.get_current_backoff_interval()
                ),
            )

    @property
    def poll_interval(self) -> float:
        """Get initial poll interval."""
        with self._attribute_lock:
            return self.backoff.initial_backoff_interval

    @poll_interval.setter
    @typechecked
    def poll_interval(self, new_interval: float) -> None:
        with self._attribute_lock:
            self.backoff.initial_backoff_interval = new_interval

    @staticmethod
    def _check_mac_address(mac_address: str) -> bool:
        pattern = r"^([0-9A-Fa-f]{2}[:-]){5}([0-9A-Fa-f]{2})$"
        return re.match(pattern, mac_address) is not None

    @property
    def broadband_lan_mac(self) -> str | None:
        """Get current gateway MAC address."""
        return self._broadband_lan_mac

    @broadband_lan_mac.setter
    @typechecked
    def broadband_lan_mac(self, new_mac: str | None) -> None:
        """Set gateway MAC address."""
        if new_mac is not None and not self._check_mac_address(new_mac):
            raise ValueError("Invalid MAC address")
        self._broadband_lan_mac = new_mac

    @property
    def device_lan_mac(self) -> str | None:
        """Get current gateway MAC address."""
        return self._device_lan_mac

    @device_lan_mac.setter
    @typechecked
    def device_lan_mac(self, new_mac: str | None) -> None:
        """Set gateway MAC address."""
        if new_mac is not None and not self._check_mac_address(new_mac):
            raise ValueError("Invalid MAC address")
        self._device_lan_mac = new_mac

    @property
    def cellular_connection_signal_strength(self) -> float:
        """Get current gateway MAC address."""
        return self._cellular_connection_signal_strength

    @cellular_connection_signal_strength.setter
    @typechecked
    def cellular_connection_signal_strength(
        self, new_signal_strength: float | None
    ) -> None:
        """Set gateway MAC address."""
        if not new_signal_strength:
            new_signal_strength = 0.0
        self._cellular_connection_signal_strength = new_signal_strength

    def set_gateway_attributes(self, gateway_attributes: dict[str, str]) -> None:
        """Set gateway attributes from dictionary.

        Args:
            gateway_attributes (dict[str,str]): dictionary of gateway attributes
        """
        for i in (
            STRING_UPDATEABLE_FIELDS
            + IPADDR_UPDATEABLE_FIELDS
            + DATETIME_UPDATEABLE_FIELDS
        ):
            temp: Any = gateway_attributes.get(i)
            if temp == "":
                temp = None
            if temp is None:
                setattr(self, i, None)
                continue
            if i in IPADDR_UPDATEABLE_FIELDS:
                try:
                    temp = ip_address(temp)
                except ValueError:
                    temp = None
            elif i in DATETIME_UPDATEABLE_FIELDS:
                try:
                    temp = int(parse_pulse_datetime(temp).timestamp())
                except ValueError:
                    temp = None
            if hasattr(self, i):
                setattr(self, i, temp)
